@@ -66,7 +66,6 @@ Point = typing.Tuple[float, float]
 Variant = typing.Any
 ModelIndexList = typing.List[int]
 QImageCleanupFunction = typing.Callable
-StringList = typing.List[str]
 
 # unfortunately, typing.Optional[t] expands to typing.Union[t, NoneType]
 # Until we can force it to create Optional[t] again, we use this.
@@ -74,45 +73,6 @@ NoneType = type(None)
 
 _S = TypeVar("_S")
 
-# Building our own Char type, which is much nicer than
-# Char = typing.Union[str, int]     # how do I model the limitation to 1 char?
-
-class _CharMeta(type):
-    def __repr__(self):
-        return '%s.%s' % (self.__module__, self.__name__)
-
-
-class Char(with_metaclass(_CharMeta)):
-    """
-    From http://doc.qt.io/qt-5/qchar.html :
-
-    In Qt, Unicode characters are 16-bit entities without any markup or
-    structure. This class represents such an entity. It is lightweight,
-    so it can be used everywhere. Most compilers treat it like an
-    unsigned short.
-
-    Here, we provide a simple implementation just to avoid long aliases.
-    """
-    __module__ = "typing"
-
-    def __init__(self, code):
-        if isinstance(code, int):
-            self.code = code & 0xffff
-        else:
-            self.code = ord(code)
-
-    def __add__(self, other):
-        return chr(self.code) + other
-
-    def __radd__(self, other):
-        return other + chr(self.code)
-
-    def __repr__(self):
-        return "typing.Char({})".format(self.code)
-
-typing.Char = Char
-
-
 MultiMap = typing.DefaultDict[str, typing.List[str]]
 
 # ulong_max is only 32 bit on windows.
@@ -184,6 +144,27 @@ class Default(_NotCalled):
 class Instance(_NotCalled):
     pass
 
+# Parameterized primitive variables
+class _Parameterized(object):
+    def __init__(self, type):
+        self.type = type
+        self.__name__ = self.__class__.__name__
+
+    def __repr__(self):
+        return "{}({})".format(
+            type(self).__name__, self.type.__name__)
+
+# Mark the primitive variables to be moved into the result.
+class ResultVariable(_Parameterized):
+    pass
+
+# Mark the primitive variables to become Sequence, Iterable or List
+# (decided in the parser).
+class ArrayLikeVariable(_Parameterized):
+    pass
+
+StringList = ArrayLikeVariable(str)
+
 
 class Reloader(object):
     """
@@ -244,27 +225,28 @@ type_map = {}
 namespace = globals()  # our module's __dict__
 
 type_map.update({
+    "...": ellipsis,
     "bool": bool,
-    "char": Char,
+    "char": int,
     "char*": str,
     "char*const": str,
     "double": float,
     "float": float,
     "int": int,
-    "List": typing.List,
+    "List": ArrayLikeVariable,
     "long": int,
     "PyCallable": typing.Callable,
     "PyObject": object,
     "PySequence": typing.Iterable,  # important for numpy
     "PyTypeObject": type,
-    "QChar": Char,
+    "QChar": str,
     "QHash": typing.Dict,
     "qint16": int,
     "qint32": int,
     "qint64": int,
     "qint8": int,
     "qintptr": int,
-    "QList": typing.List,
+    "QList": ArrayLikeVariable,
     "qlonglong": int,
     "QMap": typing.Dict,
     "QPair": typing.Tuple,
@@ -284,19 +266,30 @@ type_map.update({
     "QVector": typing.List,
     "real": float,
     "short": int,
-    "signed char": Char,
+    "signed char": int,
     "signed long": int,
+    "std.list": typing.List,
+    "std.map": typing.Dict,
+    "std.pair": typing.Tuple,
+    "std.vector": typing.List,
     "str": str,
     "true": True,
+    "Tuple": typing.Tuple,
+    "uchar": int,
+    "uchar*": str,
+    "uint": int,
+    "ulong": int,
     "ULONG_MAX": ulong_max,
-    "unsigned char": Char,
-    "unsigned int": int, # should we define an unsigned type?
+    "unsigned char": int, # 5.9
+    "unsigned char*": str,
+    "unsigned int": int,
     "unsigned long int": int, # 5.6, RHEL 6.6
     "unsigned long long": int,
     "unsigned long": int,
     "unsigned short int": int, # 5.6, RHEL 6.6
     "unsigned short": int,
-    "UnsignedShortType": int, # 5.9
+    "Unspecified": None,
+    "ushort": int,
     "void": int, # be more specific?
     "WId": WId,
     "zero(bytes)": b"",
@@ -305,7 +298,56 @@ type_map.update({
     "zero(int)": 0,
     "zero(object)": None,
     "zero(str)": "",
-    "...": "...",
+    "zero(typing.Any)": None,
+    })
+
+type_map.update({
+    # Handling variables declared as array:
+    "array double*"         : ArrayLikeVariable(float),
+    "array float*"          : ArrayLikeVariable(float),
+    "array GLint*"          : ArrayLikeVariable(int),
+    "array GLuint*"         : ArrayLikeVariable(int),
+    "array int*"            : ArrayLikeVariable(int),
+    "array long long*"      : ArrayLikeVariable(int),
+    "array long*"           : ArrayLikeVariable(int),
+    "array short*"          : ArrayLikeVariable(int),
+    "array signed char*"    : bytes,
+    "array unsigned char*"  : bytes,
+    "array unsigned int*"   : ArrayLikeVariable(int),
+    "array unsigned short*" : ArrayLikeVariable(int),
+    })
+
+type_map.update({
+    # Special cases:
+    "char*"         : bytes,
+    "QChar*"        : bytes,
+    "quint32*"      : int,        # only for QRandomGenerator
+    "quint8*"       : bytearray,  # only for QCborStreamReader and QCborValue
+    "uchar*"        : bytes,
+    "unsigned char*": bytes,
+    })
+
+type_map.update({
+    # Handling variables that are returned, eventually as Tuples:
+    "bool*"         : ResultVariable(bool),
+    "float*"        : ResultVariable(float),
+    "int*"          : ResultVariable(int),
+    "long long*"    : ResultVariable(int),
+    "long*"         : ResultVariable(int),
+    "PStr*"         : ResultVariable(str),  # module sample
+    "qint32*"       : ResultVariable(int),
+    "qint64*"       : ResultVariable(int),
+    "qreal*"        : ResultVariable(float),
+    "QString*"      : ResultVariable(str),
+    "quint16*"      : ResultVariable(int),
+    "uint*"         : ResultVariable(int),
+    "unsigned int*" : ResultVariable(int),
+    "QStringList*"  : ResultVariable(StringList),
+    })
+
+# PYSIDE-1328: We need to handle "self" explicitly.
+type_map.update({
+    "self"  : "self",
     })
 
 
@@ -329,7 +371,8 @@ def init_minimal():
 def init_sample():
     import datetime
     type_map.update({
-        "char": Char,
+        "char": int,
+        "char**": typing.List[str],
         "Complex": complex,
         "double": float,
         "Foo.HANDLE": int,
@@ -341,15 +384,16 @@ def init_sample():
         "PStr": str,
         "PyDate": datetime.date,
         "sample.bool": bool,
-        "sample.char": Char,
+        "sample.char": int,
         "sample.double": float,
         "sample.int": int,
         "sample.ObjectType": object,
         "sample.OddBool": bool,
-        "sample.Photon.TemplateBase": Missing("sample.Photon.TemplateBase"),
+        "sample.Photon.TemplateBase[Photon.DuplicatorType]": sample.Photon.ValueDuplicator,
+        "sample.Photon.TemplateBase[Photon.IdentityType]": sample.Photon.ValueIdentity,
         "sample.Point": Point,
         "sample.PStr": str,
-        "sample.unsigned char": Char,
+        "sample.unsigned char": int,
         "std.size_t": int,
         "std.string": str,
         "ZeroIn": 0,
@@ -381,6 +425,7 @@ def init_smart():
     })
     return locals()
 
+
 # The PySide Part
 def init_PySide2_QtCore():
     from PySide2.QtCore import Qt, QUrl, QDir
@@ -398,54 +443,28 @@ def init_PySide2_QtCore():
         "4294967295UL": 4294967295, # 5.6, RHEL 6.6
         "CheckIndexOption.NoOption": Instance(
             "PySide2.QtCore.QAbstractItemModel.CheckIndexOptions.NoOption"), # 5.11
+        "DescriptorType(-1)": int,  # Native handle of QSocketDescriptor
         "false": False,
         "list of QAbstractAnimation": typing.List[PySide2.QtCore.QAbstractAnimation],
         "list of QAbstractState": typing.List[PySide2.QtCore.QAbstractState],
         "long long": int,
-        "long": int,
         "NULL": None, # 5.6, MSVC
         "nullptr": None, # 5.9
         "PyByteArray": bytearray,
         "PyBytes": bytes,
-        "PyCallable": typing.Callable,
-        "PyObject": object,
-        "PySequence": typing.Iterable,  # important for numpy
-        "PySide2.QtCore.bool": bool,
-        "PySide2.QtCore.char": StringList, # A 'char **' is a list of strings.
-        "PySide2.QtCore.double": float,
-        "PySide2.QtCore.float": float,
-        "PySide2.QtCore.int": int,
-        "PySide2.QtCore.int32_t": int, # 5.9
-        "PySide2.QtCore.int64_t": int, # 5.9
-        "PySide2.QtCore.long long": int, # 5.9, MSVC 15
-        "PySide2.QtCore.long": int,
-        "PySide2.QtCore.QCborStreamReader.StringResult": typing.AnyStr,
-        "PySide2.QtCore.QChar": Char,
-        "PySide2.QtCore.qint16": int,
-        "PySide2.QtCore.qint32": int,
-        "PySide2.QtCore.qint64": int,
-        "PySide2.QtCore.qint8": int,
-        "PySide2.QtCore.qreal": float,
-        "PySide2.QtCore.QString": str,
-        "PySide2.QtCore.QStringList": StringList,
-        "PySide2.QtCore.quint16": int,
-        "PySide2.QtCore.quint32": int,
-        "PySide2.QtCore.quint64": int,
-        "PySide2.QtCore.quint8": int,
+        "QDeadlineTimer(QDeadlineTimer.Forever)": Instance("PySide2.QtCore.QDeadlineTimer"),
+        "PySide2.QtCore.QCborStreamReader.StringResult[PySide2.QtCore.QByteArray]":
+            PySide2.QtCore.QCborStringResultByteArray,
+        "PySide2.QtCore.QCborStreamReader.StringResult[QString]":
+            PySide2.QtCore.QCborStringResultString,
+        "PySide2.QtCore.QCborStreamReader.QCborStringResultByteArray":
+            PySide2.QtCore.QCborStringResultByteArray,  # 5.14, why?
+        "PySide2.QtCore.QCborStreamReader.QCborStringResultString":
+            PySide2.QtCore.QCborStringResultString,  # 5.14, why?
         "PySide2.QtCore.QUrl.ComponentFormattingOptions":
             PySide2.QtCore.QUrl.ComponentFormattingOption, # mismatch option/enum, why???
-        "PySide2.QtCore.QVariant": Variant,
-        "PySide2.QtCore.short": int,
-        "PySide2.QtCore.signed char": Char,
-        "PySide2.QtCore.uchar": Char,
-        "PySide2.QtCore.uint32_t": int, # 5.9
-        "PySide2.QtCore.unsigned char": Char, # 5.9
-        "PySide2.QtCore.unsigned int": int, # 5.9 Ubuntu
-        "PySide2.QtCore.unsigned short": int,
-        "PyTypeObject": type,
         "PyUnicode": typing.Text,
         "Q_NULLPTR": None,
-        "QChar": Char,
         "QDir.Filters(AllEntries | NoDotAndDotDot)": Instance(
             "QDir.Filters(QDir.AllEntries | QDir.NoDotAndDotDot)"),
         "QDir.SortFlags(Name | IgnoreCase)": Instance(
@@ -456,24 +475,21 @@ def init_PySide2_QtCore():
         "QGenericArgument(NULL)": ellipsis, # 5.6, MSVC
         "QGenericArgument(nullptr)": ellipsis, # 5.10
         "QGenericArgument(Q_NULLPTR)": ellipsis,
-        "QHash": typing.Dict,
         "QJsonObject": typing.Dict[str, PySide2.QtCore.QJsonValue],
         "QModelIndex()": Invalid("PySide2.QtCore.QModelIndex"), # repr is btw. very wrong, fix it?!
         "QModelIndexList": ModelIndexList,
-        "qptrdiff": int,
-        "QString": str,
+        "QModelIndexList": ModelIndexList,
         "QString()": "",
-        "QStringList": StringList,
         "QStringList()": [],
         "QStringRef": str,
-        "Qt.HANDLE": int, # be more explicit with some consts?
-        "quintptr": int,
+        "QStringRef": str,
+        "Qt.HANDLE": int, # be more explicit with some constants?
         "QUrl.FormattingOptions(PrettyDecoded)": Instance(
             "QUrl.FormattingOptions(QUrl.PrettyDecoded)"),
-        "QVariant": Variant,
         "QVariant()": Invalid(Variant),
         "QVariant.Type": type, # not so sure here...
         "QVariantMap": typing.Dict[str, Variant],
+        "QVariantMap": typing.Dict[str, Variant],
     })
     try:
         type_map.update({
@@ -493,16 +509,12 @@ def init_PySide2_QtGui():
         "GL_COLOR_BUFFER_BIT": GL_COLOR_BUFFER_BIT,
         "GL_NEAREST": GL_NEAREST,
         "int32_t": int,
-        "PySide2.QtCore.uint8_t": int, # macOS 5.9
-        "PySide2.QtGui.QGenericMatrix": Missing("PySide2.QtGui.QGenericMatrix"),
-        "PySide2.QtGui.QPlatformSurface": int, # a handle
-        "QList< QTouchEvent.TouchPoint >()": [], # XXX improve?
         "QPixmap()": Default("PySide2.QtGui.QPixmap"), # can't create without qApp
+        "QPlatformSurface*": int, # a handle
         "QVector< QTextLayout.FormatRange >()": [], # do we need more structure?
         "uint32_t": int,
         "uint8_t": int,
         "USHRT_MAX": ushort_max,
-        "WId": WId,
     })
     return locals()
 
@@ -513,7 +525,6 @@ def init_PySide2_QtWidgets():
     type_map.update({
         "QMessageBox.StandardButtons(Yes | No)": Instance(
             "QMessageBox.StandardButtons(QMessageBox.Yes | QMessageBox.No)"),
-        "QVector< int >()": [],
         "QWidget.RenderFlags(DrawWindowBackground | DrawChildren)": Instance(
             "QWidget.RenderFlags(QWidget.DrawWindowBackground | QWidget.DrawChildren)"),
         "SH_Default": QStyleHintReturn.SH_Default,
@@ -536,9 +547,17 @@ def init_PySide2_QtSql():
 
 
 def init_PySide2_QtNetwork():
+    from PySide2.QtNetwork import QNetworkRequest
+    best_structure = typing.OrderedDict if getattr(typing, "OrderedDict", None) else typing.Dict
     type_map.update({
-        "QMultiMap": MultiMap,
+        "QMultiMap[PySide2.QtNetwork.QSsl.AlternativeNameEntryType, QString]":
+            best_structure[PySide2.QtNetwork.QSsl.AlternativeNameEntryType, typing.List[str]],
+        "DefaultTransferTimeoutConstant":
+            QNetworkRequest.TransferTimeoutConstant,
+        "QNetworkRequest.DefaultTransferTimeoutConstant":
+            QNetworkRequest.TransferTimeoutConstant,
     })
+    del best_structure
     return locals()
 
 
@@ -557,6 +576,7 @@ def init_PySide2_QtMultimedia():
     check_module(PySide2.QtMultimediaWidgets)
     type_map.update({
         "QGraphicsVideoItem": PySide2.QtMultimediaWidgets.QGraphicsVideoItem,
+        "qint64": int,
         "QVideoWidget": PySide2.QtMultimediaWidgets.QVideoWidget,
     })
     return locals()
@@ -569,26 +589,23 @@ def init_PySide2_QtOpenGL():
         "GLfloat": float, # 5.6, MSVC 15
         "GLint": int,
         "GLuint": int,
-        "PySide2.QtOpenGL.GLint": int,
-        "PySide2.QtOpenGL.GLuint": int,
     })
     return locals()
 
 
 def init_PySide2_QtQml():
     type_map.update({
-        "PySide2.QtQml.bool volatile": bool,
         "QJSValueList()": [],
-        "QVariantHash()": typing.Dict[str, Variant],  # XXX sorted?
+        "QVariantHash()": typing.Dict[str, Variant],    # from 5.9
     })
     return locals()
 
 
 def init_PySide2_QtQuick():
     type_map.update({
-        "PySide2.QtCore.uint": int,
-        "PySide2.QtQuick.QSharedPointer": int,
-        "T": int,
+        "PySide2.QtQuick.QSharedPointer[PySide2.QtQuick.QQuickItemGrabResult]":
+            PySide2.QtQuick.QQuickItemGrabResult,
+        "UnsignedShortType": int,
     })
     return locals()
 
@@ -602,6 +619,7 @@ def init_PySide2_QtScript():
 
 def init_PySide2_QtTest():
     type_map.update({
+        "PySide2.QtTest.QTest.PySideQTouchEventSequence": PySide2.QtTest.QTest.QTouchEventSequence,
         "PySide2.QtTest.QTouchEventSequence": PySide2.QtTest.QTest.QTouchEventSequence,
     })
     return locals()
@@ -622,6 +640,10 @@ def init_PySide2_QtDataVisualization():
     QtDataVisualization.QSurfaceDataArray = typing.List[QtDataVisualization.QSurfaceDataRow]
     type_map.update({
         "100.0f": 100.0,
+        "QtDataVisualization.QBarDataArray": QtDataVisualization.QBarDataArray,
+        "QtDataVisualization.QBarDataArray*": QtDataVisualization.QBarDataArray,
+        "QtDataVisualization.QSurfaceDataArray": QtDataVisualization.QSurfaceDataArray,
+        "QtDataVisualization.QSurfaceDataArray*": QtDataVisualization.QSurfaceDataArray,
     })
     return locals()
 
